-- Command menu. 
-- Displays command in a similar fashion as the main Turbine menu.
-- Commands are displayed bottom/up if the menu is opened in the bottom half of the display

-- HACK: If displaying bottom/up, the menu will shift upward as if SetTop() is actually SetBottom(). 

-- TODO: handle icons
-- TODO: Whether the menu is displaying bottom/up or top/down should be externalized

import "Turbine.UI";
import "Equendil.UI.DecoratedContainer";
import "Equendil.Utils";

-- resources
local GradiantBlue = 0x41101B27;
local GradiantBlueReversed = 0x41101D9B;

-- Individual controls displayed in the menu
local MenuItem = class( Turbine.UI.Control );
function MenuItem:Constructor( command )
	Turbine.UI.Control.Constructor( self );
	
	self.command = command;
	self.Clicked = Equendil.Utils.WeakEvent();
	self.mouseInside = false;
	
	self:SetSize( 240, 33 );
	self:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	self:SetBackground( GradiantBlue );

--[[	No icons for now
	self.icon = Turbine.UI.Control();
	self.icon:SetParent( self );
	self.icon:SetSize( 25, 32 );
	self.icon:SetBlendMode( Turbine.UI.BlendMode.Normal );
	self.icon:SetMouseVisible( false );
	if self.command:GetIcon() ~= nil then
		self.icon:SetBackground( self.command:GetBackdrop() );
	end
--]]
	
	-- setup a big centered label for the command name
	self.label = Turbine.UI.Label();
	self.label:SetParent( self );
	self.label:SetSize(240,32); -- we use a height one pixel less to leave a separation line between items from the gradiant.
	self.label:SetPosition( 0, 0 );
	self.label:SetFont( Turbine.UI.Lotro.Font.TrajanPro16 );
	self.label:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleCenter );
	self.label:SetFontStyle( Turbine.UI.FontStyle.Outline );
	self.label:SetForeColor( Equendil.UI.Color.White );
	self.label:SetMouseVisible( false );
	self.label:SetText( command:GetName() );
	self.label:SetBackColor( Equendil.UI.Color.Black ); -- cover the gradient save for 1 pixel
	self.label:SetBackColorBlendMode( Turbine.UI.BlendMode.Overlay ); 

	self:SetMouseVisible( true );
	
	-- mouse events
	self.MouseEnter = function( sender, args ) 
		self.mouseInside = true; -- we want to remember where we are for MouseUp events
		self.label:SetBackColor( Equendil.UI.Color.Transparent ); -- Label is made transparent to uncover the gradient behind
	end
	self.MouseLeave = function( sender, args ) 
		self.label:SetBackColor( Equendil.UI.Color.Black ); -- label made solid to hide the gradient behind (save for one pixel for separation)
		self.mouseInside = false;
	end
	self.MouseDown = function( sender, args )
		self:SetBackground( GradiantBlueReversed );
	end
	self.MouseUp = function( sender, args )
		self:SetBackground( GradiantBlue );
		if self.mouseInside then
			self.Clicked( args.Button ); -- fire a click event if the mouse is still inside the control
		end
	end
	
	-- setup a callback to update when the command changes
	self.updateAction = function() self:_Update(); end
	self.command.Changed:Add(self.updateAction); 
end

-- Garbage collector friendly 'destructor'
function MenuItem:Dispose()
	self:SetParent( nil );
	self.Clicked = nil;
	self.command.Changed:Remove( self.updateAction );
end

-- update the menu item 
function MenuItem:_Update()
	self.label:SetText( self.command:GetName() );
end

-- run the command therein
function MenuItem:Execute()
	self.command:Execute();
end

-- returns whether a menu item should be displayed
function MenuItem:IsHidden()
	return self.command:GetMode() == Equendil.PluginMng.Command.Mode.RunOnly and not self.command:GetPlugin():IsLoaded();
end

-- Main class
Menu = class( Turbine.UI.Window );
function Menu:Constructor()
	Turbine.UI.Window.Constructor( self );
	
	self.commands = {};
	self.menuItems = {};
	
	self:SetVisible( false );
	self:SetSize( 250, 33 );
	-- self:SetBackColor( Equendil.UI.Color.Black );
	
	self.borders = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.RoundSilver );
	self.borders:SetParent( self );
	self.borders:SetZOrder( 100 ); -- arbitrarily large value to keep the borders on top of the menu items so there's no nasty clipping
	self.borders:SetSize( self:GetSize() );
end

-- set the list of commands to manage
function Menu:Set( commands )
	-- get rid of the old menu items
	for k, menuItem in pairs( self.menuItems ) do
		menuItem:Dispose();
	end
	self.menuItems = {};

	-- and build new ones
	self.commands = commands;
	for i, command in ipairs( commands ) do
		local menuItem = MenuItem( command );
		menuItem:SetParent( self );
		menuItem:SetLeft( 5 );
		menuItem:SetVisible( false );
		
		-- on click, execute the command and hide the menu. 
		-- We should probably fire an event instead for someone else to deal with the command, but let's keep this simple
		menuItem.Clicked = function( button )
			menuItem:Execute();
			self:Show(false);
		end
		table.insert( self.menuItems, menuItem );
	end

end	

-- Show/Hide the menu
function Menu:Show( flag )
	if not flag then 
		-- hide the menu
		self:Close();
		return;
	end
	
	-- find how many items need to be displayed
	local numVisible = 0;
	for i, menuItem in ipairs( self.menuItems ) do
		if not menuItem:IsHidden() then
			numVisible = numVisible + 1;
		end
	end

	-- set the height accordingly
	local height = 10 + 33 * numVisible;
	self:SetHeight( height );
	self.borders:SetHeight( height );
	
	-- initialize position and offset depending if we draw the menu top down or bottom up
	-- which is determined by the quadrant on the display in which we're displaying
	-- position of the menu is adjusted if displaying bottom up
	local topItem = 5;
	local offset = 33;
	if self:GetTop() > Turbine.UI.Display.GetHeight() / 2 then 
		-- menu is below the top half of the display, we'll display items bottom up
		offset = -33;
		topItem = height - 5 - 33;
		self:SetTop( self:GetTop() - height );
	end
	
	-- position items in the menu
	for i, menuItem in ipairs( self.menuItems ) do
		if not menuItem:IsHidden() then
			menuItem:SetTop( topItem );
			menuItem:SetVisible( true );
			topItem = topItem + offset;
		else
			menuItem:SetVisible( false );
		end
	end

	-- open the menu proper
	self:SetVisible( true );
	self:Activate();
end
