import "Turbine.UI";

-- A VirtualListBox is a list box that doesn't handle its own data, instead requesting controls that need to be displayed on the fly.
-- This presents two important advantages
-- 1) Decoupling of data & user interface. UI elements shouldn't be handling data that is not immediately relevant to display or user interaction.
-- 2) No overhead in handling large lists, ie it's fast

-- NOTE: If selection mode is not set to NoSelection, events on item controls might be captured by this control
-- NOTE: selection highlightning is handled by callin the SetSelected( true/false) method on an item
-- NOTE: Does not handle variable height items (pain in the bottom).. Set the item height prior to using the control but after sizing it 

-- TODO: handle setting the item height properly
-- TODO: handle resizing
-- TODO: multiple selection entirely untested
-- TODO: horizontal scrollbar

SelectionMode = { SingleSelection = 1, MultipleSelection = 2, NoSelection = 3 };

VirtualListBox = class( Turbine.UI.Control );

function VirtualListBox:Constructor( getItemFunction )
	Turbine.UI.Control.Constructor( self );

	-- GetItemControl( n, selectionStatus ) is called for each item the listbox needs to draw.
	-- It should return a control for the nth item in the list
	-- selectionStatus = true if the item is selected, false or nil otherwise
	self.GetItemControl = getItemFunction; 

	-- event when selectiont has changed, no paramaters
	self.SelectionChanged = Equendil.Utils.Event();
	
	self.itemCount = 0;
	self.position = 1; -- position in the list
	self.numItemsDisplayed = 0;
	self.multiSelection = false;
	self.selectionMode = SelectionMode.SingleSelection;
	self.itemHeight = itemHeight;
	
	self.selectedIndices = {};
	self.lastSelectedIndex = -1;

	self.verticalScrollBar = nil;
	
	self.itemControls = {};
end

-- set the height of any item that will be displayed in the list box, we do not handle variable height items
function VirtualListBox:SetItemHeight( itemHeight )
	self.itemHeight = itemHeight;
	self.numItemsDisplayed = math.floor( self:GetHeight() /  self.itemHeight );
end

-- Set the selection mode of the list box, one of the SelectionMode enum
function VirtualListBox:SetSelectionMode( selectionMode )
	self.selectionMode = selectionMode;
	
	if self.selectionMode ~= SelectionMode.MultipleSelection then
		self.selectedIndices = {};
	end
	
	if self.selectionMode == SelectionMode.NoSelection then
		self.lastSelectedIndex = -1;
	end
end

-- get the selection mode of the listbox
function VirtualListBox:GetSelectionMode()
	return self.selectionMode;
end

-- get the position within the list
function VirtualListBox:GetListPosition()
	return self.position;
end

-- set the position within the list
function VirtualListBox:SetListPosition( position )
	self.position = position;
	self:Refresh();
end

-- Ensure the nth item is visible in the listbox. If not visible it is brought to the top or bottom of the list depending on current position
function VirtualListBox:MakeVisible( n )
	if n < self.position then
		-- position is before the first item displayed, bring to top
		self.position = n; 
		self:Refresh();
	elseif n > self.position + self.numItemsDisplayed - 1 then
		-- position is after the last item displayed, bring it to bottom
		self.position = self:_MaxPosition( n );
		
		self:Refresh();
	else
		return; -- already visible
	end

	-- update the scroll bar accordingly
	if self.verticalScrollBar ~= nil then
		-- remove the event handler before we set the value, we don't want self generated events
		local eventHandler = self.verticalScrollBar.ValueChanged;
		self.verticalScrollBar.ValueChanged = nil;
		self.verticalScrollBar:SetValue( self.position );
		self.verticalScrollBar.ValueChanged = eventHandler;
	end
end

-- find what the position should be set so that the nth item is at the bottom. If there are too few items to display position is top of the list
function VirtualListBox:_MaxPosition( n )
	return math.max( 1, n - self.numItemsDisplayed + 1 );
end

-- set how many items are in the listbox, will cause loss of selection and refresh
function VirtualListBox:SetItemCount( n )
	self.itemCount = n;
	local maxPosition = self:_MaxPosition( n );
	if self.position > maxPosition then
		self.position = maxPosition;
	end
	-- get rid of old selection
	self.lastSelectedIndex = -1;
	self.selectedIndices = {};
	
	self:Refresh();
	self:_UpdateScrollBars();
	self.SelectionChanged();
end

-- Return the number of items in the listbox
function VirtualListBox:GetItemCount( n )
	return self.itemCount;
end

-- Clear all controls currently displayed
function VirtualListBox:_ClearItems()
	-- clear the item control list
	for i, control in ipairs( self.itemControls ) do
		control:SetParent( nil );
		-- clear callbacks if necessary
		if self.selectionMode ~= SelectionMode.NoSelection then
			control.MouseUp = nil; -- TODO: we might have zombie callbacks here if the selection mode is changed after initialization of the listbox
		end
	end
	self.itemControls = {};
end

-- Redraw the list
function VirtualListBox:Refresh()
	self:_ClearItems();
	
	-- get & position controls
	local top = 0;
	for i = self.position, self.itemCount do
		-- get the control to display
		local control = nil;
		if self.selectionMode == SelectionMode.MultipleSelection then
			control = self.GetItemControl( i, self.selectedIndices[i] );
		elseif self.selectionMode == SelectionMode.SingleSelection then
			control = self.GetItemControl( i, self.lastSelectedIndex == i );
		end
		
		-- and set it up
		control:SetParent( self );
		control:SetPosition( 0, top );
		if self.selectionMode ~= SelectionMode.NoSelection then
			control:SetMouseVisible( true );
			control.MouseUp = function( sender, args )
				-- Turbine.Log:Information( "selected " .. tostring( i ) );
				if self.selectionMode == SelectionMode.MultipleSelection then
					self:_ToggleSelection( i );
				else
					self:Select( i );
				end
			end
		end

		table.insert( self.itemControls, control );
		
		-- if we're beyond the bottom, stop displaying controls
		top = top + control:GetHeight();
		if top >= self:GetHeight() then
			break;
		end
	end
end

-- Clear any and all selection
function VirtualListBox:ClearSelection()
	self.lastSelectedIndex = -1;
	self.selectedIndices = {};
end

-- returns the first selected index
function VirtualListBox:GetSelectedIndex()
	return self.lastSelectedIndex;
end

-- returns a table of all selected indices
function VirtualListBox:GetSelectedIndices()
	return self.selectedIndices;
end

-- returns the nth item to display provided it's visible (ie, being currently handled)
function VirtualListBox:_GetNthItem( n )
	if n >= self.position and n < self.position + self.numItemsDisplayed then
		return self.itemControls[n - self.position + 1];
	end
	
	return nil;
end

-- Set selection of the nth item to <flag>. If flag is nil then toggle the selection of that item.
-- External method
function VirtualListBox:Select( n )
	if self.selectionMode == SelectionMode.NoSelection then	return end

	if self.selectionMode == SelectionMode.MultipleSelection then
		self.selectedIndices[n] = true;
	else 
		local oldSelectedItem = self:_GetNthItem( self.lastSelectedIndex );
		if oldSelectedItem ~= nil then
			oldSelectedItem:SetSelected( false );
		end
	end

	self.lastSelectedIndex = n;
	local selectedItem = self:_GetNthItem(n);
	if selectedItem ~= nil then
		selectedItem:SetSelected( true );
	end
		
	self.SelectionChanged();
end

-- Toggle the selection of an item
-- Private method to react to user actions
function VirtualListBox:_ToggleSelection( n )
	if self.selectionMode ~= SelectionMode.MultipleSelection then return end

	if self.selectedIndices[n] then
		self.selectedIndices[n] = nil;
		self:_GetNthItem(n):SetSelected( false );
	else
		self.selectedIndices[n] = true;
		self:_GetNthItem(n):SetSelected( true );
	end

	-- if we've just deselected the last selected index, set it to the first selected index if any
	if self.lastSelectedIndex == n then
		self.lastSelectedIndex = -1;
		for i, v in ipairs(self.selectedIndices) do
			self.lastSelectedIndex = i;
			break;
		end
	end
end

-- Update the configuration of the scroll bars
function VirtualListBox:_UpdateScrollBars()
	-- if all the items are displayed make the scrollbar invisible
	if self.verticalScrollBar ~= nil then
		if self.numItemsDisplayed >= self.itemCount then
			self.verticalScrollBar:SetVisible( false );
		else
			self.verticalScrollBar:SetVisible( true );
			local maxPosition = self:_MaxPosition( self.itemCount );
			
			-- the "Large Change" value determines how high the thumb button will be, no method to set its height unfortunately, so it tries to "guess" ....
			-- needs to be set as a ratio of the max value to make its height a ratio of the height of the scrollbar
			-- It does not seem to handle non integer values though, so not perfect
			-- Also makes it not terribly useful as a "large change" action ...
			local largeChange = ( self.numItemsDisplayed * maxPosition ) / self.itemCount; --division by 0 possible if the list isn't tall enough to display one item ...
			self.verticalScrollBar:SetLargeChange( largeChange );

			-- SetMaximum has to be called after SetLargeChange or it doesn't seem to be working
			if maxPosition == 2 then
				self.verticalScrollBar:SetMaximum( 3 ); -- WORKAROUND: bloody scrollbar doesn't seem to handle a range of 2 values or i don't understand how it works, either way ...
			else
				self.verticalScrollBar:SetMaximum( maxPosition );
			end
		end
	end
	
	 -- TODO: no horizontal scroll bar for now.
end

-- Set a vertical scroll bar to the listbox
function VirtualListBox:SetVerticalScrollBar( scrollbar )
	self.verticalScrollBar = scrollbar;
	self.verticalScrollBar:SetMinimum( 1 );
	self.verticalScrollBar:SetSmallChange( 1 );

	self:_UpdateScrollBars();
	
	self.verticalScrollBar.ValueChanged = function( sender, args )
		self.position = math.floor( self.verticalScrollBar:GetValue() / 1 );
		self:Refresh();
	end
end

-- Set an horizontal scroll bar to the listbox
function VirtualListBox:SetHorizontalScrollBar( scrollbar )
	-- TODO
end
